# -*- coding: utf-8 -*-
# @Time    : 2022/10/23 19:00
# @Author  : mf.liang
# @File    : dhcp_pkt_v6.py
# @Software: PyCharm
# @desc    :
from argparse import Namespace
from scapy.layers.dhcp import BOOTP, DHCP
from dhcptool.base_dhcp_pkt import BasePkt
from dhcptool.env_args import pkt_result
from dhcptool.options import Dhcp4Options
from dhcptool.tools import Tools


class Dhcp4Pkt(BasePkt):

    def __init__(self, args: Namespace) -> None:
        super(Dhcp4Pkt, self).__init__(args)
        self.udp.sport = 67
        self.udp.dport = 67

        self.ip.src = self.args.ip_src or Tools.get_local_address().get('ipv4') or '0.0.0.0'
        self.ip.dst = self.args.dhcp_server or '255.255.255.255'
        if self.args.broadcast:
            self.ether.dst = 'ff:ff:ff:ff:ff:ff'
            self.ip.src = '0.0.0.0'
            self.ip.dst = '255.255.255.255'
        self.bootp.giaddr = self.args.relay_forward or self.args.ip_src or Tools.get_local_address().get('ipv4')
        self.ether_ip_udp_bootp = self.ether / self.ip / self.udp / self.bootp
        self.make_options = Dhcp4Options(self.args)
        self.options_list = self.make_options.make_options_list()

    def make_pkts(self, message_type, **kwargs) -> DHCP:
        """
        组装  discover/request/decline/inform报文，
        :param message_type: 请求类型
        :return:
        """
        options = [("message-type", message_type)]
        if not self.args.single and message_type != 'discover':
            if message_type == 'request':
                response_pkt = pkt_result.get('dhcp4_offer').get(timeout=self.timeout)
                if kwargs.get('message') == 'exception_request':
                    response_pkt = BOOTP(yiaddr='0.0.0.1')
            else:
                response_pkt = pkt_result.get('dhcp4_ack').get(timeout=self.timeout)

            yiaddr = response_pkt[BOOTP].yiaddr
            options.append(("requested_addr", yiaddr))

        [options.append(i) for i in self.options_list]
        make_pkt = self.ether_ip_udp_bootp / DHCP(options=options)
        return make_pkt

    def dhcp4_discover(self) -> DHCP:
        """
        制作 discover包
        :return:
        """
        discover_pkt = self.make_pkts("discover")
        return discover_pkt

    def dhcp4_offer(self) -> DHCP:
        pass

    def dhcp4_request(self) -> DHCP:
        """
        制作 request包
        :return:
        """
        request_pkt = self.make_pkts('request')
        return request_pkt

    def dhcp4_exception_request(self) -> DHCP:
        """
        制作 一个会回nak的request包
        :return:
        """
        request_pkt = self.make_pkts('request', message='exception_request')
        return request_pkt

    def dhcp4_ack(self) -> DHCP:
        pass

    def dhcp4_decline(self) -> DHCP:
        """
        制作 decline包
        :return:
        """
        decline_pkt = self.make_pkts('decline')
        return decline_pkt

    def dhcp4_release(self) -> DHCP:
        """
        制作 release包
        :return:
        """
        release_pkt = self.make_pkts('release')
        return release_pkt

    def dhcp4_inform(self) -> DHCP:
        """
        制作 inform包
        :return:
        """
        inform_pkt = self.make_pkts('inform')
        return inform_pkt



# -*- coding: utf-8 -*-
# @Time    : 2022/10/23 19:00
# @Author  : mf.liang
# @File    : dhcp_pkt_v6.py
# @Software: PyCharm
# @desc    :
import copy
import uuid
from scapy.layers.dhcp6 import (
    DHCP6OptClientId,
    DHCP6OptIA_NA,
    DHCP6OptIA_PD,
    DHCP6OptServerId,
    DHCP6_RelayForward,
    DUID_LLT,
    DUID_LL,
    DUID_UUID,
    DHCP6OptRelayMsg,
    DHCP6OptIfaceId,
    DUID_EN,
    DHCP6, All_DHCP_Relay_Agents_and_Servers, DHCP6OptSubscriberID, DHCP6OptIA_TA
)
from dhcptool.base_dhcp_pkt import BasePkt
from dhcptool.env_args import pkt_result
from dhcptool.options import Dhcp6Options
from typing import Union

from dhcptool.tools import Tools


class Dhcp6Pkt(BasePkt):

    def __init__(self, args) -> None:
        super(Dhcp6Pkt, self).__init__(args)
        self.ipv6.src = self.args.ip_src or Tools.get_local_address().get('ipv6')
        self.ipv6.src = self.args.ip_src

        # TODO: 同网段下 使用 单播，和异网段下使用单播，可以尝试去掉或加上这个试试，具体效果没有尝试
        if not self.args.broadcast:
            self.ipv6.dst = self.args.dhcp_server or All_DHCP_Relay_Agents_and_Servers

        self.ether_ipv6_udp = self.ether / self.ipv6 / self.udp
        self.duid = DUID_LLT(lladdr=self.mac, timeval=self.xid)
        if self.args.duid:
            if self.args.duid == 'llt':
                self.duid = DUID_LLT(lladdr=self.mac, timeval=self.xid)
            elif self.args.duid == 'll':
                self.duid = DUID_LL(lladdr=self.mac)
            elif self.args.duid == 'en':
                self.duid = DUID_EN(enterprisenum=1058949886, id=bytes.fromhex('7db44f928a1d4ef7c842c434'))
            elif self.args.duid == 'uuid':
                self.duid = DUID_UUID(uuid=uuid.uuid4())
            else:
                self.duid = bytes.fromhex(self.args.duid)
        self.opt_client_id = DHCP6OptClientId(duid=self.duid)
        self.duid_en = DUID_EN(enterprisenum=1058949886, id=bytes.fromhex('7db44f928a1d4ef7c842c434'))
        self.duid_xid = Tools.get_xid_by_duid(self.duid)
        self.opt_ia_na = DHCP6OptIA_NA(iaid=self.duid_xid)
        self.opt_ia_ta = DHCP6OptIA_TA(iaid=self.duid_xid)
        self.opt_ia_pd = DHCP6OptIA_PD(iaid=self.duid_xid)
        self.opt_server_id = DHCP6OptServerId(duid=self.duid_en)
        self.relay_forward = DHCP6_RelayForward(linkaddr=self.args.relay_forward)
        self.make_options = Dhcp6Options(self.args)
        self.options_list, self.parent_options_list = self.make_options.make_options_list()
        self.dhcp6_options = self.make_options.parse_options()

    def make_pkts(self, message_type) -> DHCP6:
        """
        解析并制作报文
        :param message_type: 报文消息类型
        :return:
        """
        opt_client_id = self.opt_client_id / self.opt_server_id
        if self.args.single and self.dhcp6_options:
            for option in self.dhcp6_options:
                if int(option[0]) == 5:
                    initial_length = len(opt_client_id)
                    if self.args.na:
                        self.opt_ia_na.ianaopts.append(self.make_options.sub_option_5(addr=option[1]))
                        opt_client_id.add_payload(self.opt_ia_na)
                    if self.args.ta:
                        self.opt_ia_ta.iataopts.append(self.make_options.sub_option_5(addr=option[1]))
                        opt_client_id.add_payload(self.opt_ia_ta)
                    if len(opt_client_id) == initial_length:
                        self.opt_ia_na.ianaopts.append(self.make_options.sub_option_5(addr=option[1]))
                        opt_client_id.add_payload(self.opt_ia_na)
                if int(option[0]) == 26:
                    self.opt_ia_pd.iapdopt.append(self.make_options.sub_option_26(prefix=option[1]))
                    opt_client_id.add_payload(self.opt_ia_pd)
        else:
            if message_type == 'request':
                reply_pkt = pkt_result.get('dhcp6_advertise').get(timeout=self.timeout)

            else:
                reply_pkt = pkt_result.get('dhcp6_reply').get(timeout=self.timeout)
            opt_client_id = reply_pkt[DHCP6OptClientId]

        pkt = self.ether_ipv6_udp / self.dhcp6 / opt_client_id / self.options_list
        return pkt

    def dhcp6_solicit(self) -> Union[DHCP6, DHCP6_RelayForward]:
        """
        制作solicit包
        :return:
        """
        self.dhcp6.msgtype = 'SOLICIT'
        if self.dhcp6_options:
            for option in self.dhcp6_options:
                if int(option[0]) == 5:
                    update_status = False
                    if self.args.na:
                        update_status = True
                        self.opt_ia_na.ianaopts.append(self.make_options.sub_option_5(addr=option[1]))
                    if self.args.ta:
                        update_status = True
                        self.opt_ia_ta.iataopts.append(self.make_options.sub_option_5(addr=option[1]))
                    if not update_status:
                        self.opt_ia_na.ianaopts.append(self.make_options.sub_option_5(addr=option[1]))
                elif int(option[0]) == 26:
                    self.opt_ia_pd.iapdopt.append(self.make_options.sub_option_26(prefix=option[1]))
        # 拼接数据包
        solicit_pkt = self.ether_ipv6_udp / self.dhcp6 / self.opt_client_id
        initial_length = len(solicit_pkt)
        if self.args.na:    # NA
            solicit_pkt.add_payload(self.opt_ia_na)
        if self.args.ta:    # TA
            solicit_pkt.add_payload(self.opt_ia_ta)
        if self.args.pd:    # PD
            solicit_pkt.add_payload(self.opt_ia_pd)
        if len(solicit_pkt) == initial_length:
            solicit_pkt.add_payload(self.opt_ia_na)
        solicit_pkt.add_payload(self.options_list)
        if self.args.broadcast is False:
            solicit_pkt = self.dhcp6_relay_ward(self.parent_options_list, solicit_pkt[DHCP6])
            return solicit_pkt
        return solicit_pkt

    def dhcp6_advertise(self) -> Union[DHCP6, DHCP6_RelayForward]:
        pass

    def dhcp6_request(self) -> Union[DHCP6, DHCP6_RelayForward]:
        """
        制作request包
        :return:
        """
        self.dhcp6.msgtype = 'REQUEST'
        request_pkt = self.make_pkts('request')
        if self.args.broadcast is False:
            request_pkt = self.dhcp6_relay_ward(self.parent_options_list, request_pkt[DHCP6])
            return request_pkt
        return request_pkt

    def dhcp6_reply(self) -> DHCP6:
        pass

    def dhcp6_renew(self) -> Union[DHCP6, DHCP6_RelayForward]:
        """
        制作renew包
        :return:
        """
        self.dhcp6.msgtype = 'RENEW'
        renew_pkt = self.make_pkts('renew')
        if self.args.broadcast is False:
            renew_pkt = self.dhcp6_relay_ward(self.parent_options_list, renew_pkt[DHCP6])
            return renew_pkt
        return renew_pkt

    def dhcp6_release(self) -> Union[DHCP6, DHCP6_RelayForward]:
        """
        制作release包
        :return:
        """
        self.dhcp6.msgtype = 'RELEASE'
        release_pkt = self.make_pkts('release')
        if self.args.broadcast is False:
            release_pkt = self.dhcp6_relay_ward(self.parent_options_list, release_pkt[DHCP6])
            return release_pkt
        return release_pkt

    def dhcp6_decline(self) -> Union[DHCP6, DHCP6_RelayForward]:
        """
        制作decline包
        :return:
        """
        self.dhcp6.msgtype = 'DECLINE'
        decline_pkt = self.make_pkts('decline')
        if self.args.broadcast is False:
            decline_pkt = self.dhcp6_relay_ward(self.parent_options_list, decline_pkt[DHCP6])
            return decline_pkt
        return decline_pkt

    def dhcp6_relay_ward(self,parent_options=None, pkt=None) -> DHCP6_RelayForward:
        """
        制作中继包
        :return:
        """
        relay_forward_pkt = self.ether_ipv6_udp / self.relay_forward
        if parent_options:
            relay_forward_pkt = relay_forward_pkt / parent_options
        if pkt:
            relay_forward_pkt.add_payload(DHCP6OptRelayMsg(message=pkt))
        return relay_forward_pkt



# -*- coding: utf-8 -*-
# @Time    : 2022/10/23 18:56
# @Author  : mf.liang
# @File    : tools.py
# @Software: PyCharm
# @desc    :

import hashlib
import ipaddress
import platform
import re
from argparse import Namespace
from inspect import getmodule, stack
from typing import Optional
import yaml
from scapy.compat import raw
from scapy.config import conf
from scapy.interfaces import get_working_if, get_if_list
from scapy.arch import get_if_hwaddr
from scapy.layers.dhcp import DHCPTypes, DHCP, BOOTP
from scapy.layers.dhcp6 import dhcp6types, DHCP6OptIAAddress, DHCP6OptRelayMsg, DHCP6OptIAPrefix, DUID_LLT
from scapy.layers.inet import IP, UDP
from scapy.layers.inet6 import IPv6
from scapy.layers.l2 import Ether
from scapy.utils import mac2str, str2mac
from scapy.volatile import RandMAC
from dhcptool.env_args import pkt_result, logs, summary_result, global_var
import time


class Tools:

    @staticmethod
    def mac_self_incrementing_or_subtracting(mac, num, mac_type, offset=1) -> str:
        """
        mac自增
        :param mac_type:
        :param num:
        :param offset:
        :return:
        """
        mac = ''.join(mac.split(':'))
        #  使用format格式化字符串，int函数，按照16进制算法，将输入的mac地址转换成十进制，然后加上偏移量
        # {:012X}将十进制数字，按照16进制输出。其中12表示只取12位，0表示不足的位数在左侧补0
        if mac_type == 'ASC':
            mac_address = "{:012X}".format(int(mac, 16) + offset * num)
        else:
            mac_address = "{:012X}".format(int(mac, 16) - offset * num)
        mac_address = ':'.join(re.findall('.{2}', mac_address)).lower()
        return mac_address

    @staticmethod
    def get_mac(args: Namespace = None) -> bytes:
        """
        获取mac信息
        :return:
        """
        ip_type = Tools.check_ip_version(args.dhcp_server)
        if args.mac and args.fixe:
            mac = mac2str(args.mac)
        elif args.mac and not args.fixe:
            mac = Tools.mac_self_incrementing_or_subtracting(args.mac, global_var.get('tag'), 'ASC')
            mac = mac2str(mac)
        elif not args.mac and args.broadcast and ip_type == 'ipv4':
            mac = get_if_hwaddr(get_working_if().name)
            mac = Tools.mac_self_incrementing_or_subtracting(mac, global_var.get('tag'), 'ASC')
            mac = mac2str(mac)
        else:
            mac = mac2str(RandMAC())
        global_var.update({"generate_mac": mac})
        return mac

    @staticmethod
    def get_xid_by_mac(mac) -> int:
        """
        根据mac生成hash
        :return:
        """
        mac = str2mac(mac).encode('utf-8')
        m = hashlib.md5()
        m.update(mac)
        mac_xid = int(str(int(m.hexdigest(), 16))[0:9])
        return mac_xid

    @staticmethod
    def get_xid_by_duid(duid: DUID_LLT) -> int:
        """
        根据mac生成hash
        :return:
        """
        duid = raw(duid)
        m = hashlib.md5()
        m.update(duid)
        duid_xid = int(str(int(m.hexdigest(), 16))[0:9])
        return duid_xid

    @staticmethod
    def convert_code(data) -> hex:
        """
        字节/16进制相互转换
        :param data:
        :return:
        """
        if isinstance(data, bytes):  # 转 16进制
            data = data.hex()
        else:  # 字符串转化成字节码
            data = bytes.fromhex(data)
        return data

    @staticmethod
    def get_local_address():
        # 获取网络接口信息
        interfaces = [interface for interface in get_if_list() if interface.startswith('e')]
        # 遍历网络接口，查找IP地址
        try:
            for interface in interfaces:
                # 获取网络接口的详细信息
                ipv4_info = conf.ifaces[interface].ips[4]
                ipv6_info = conf.ifaces[interface].ips[6]
                ipv4_address = [i for i in ipv4_info if not i.startswith('127.')]
                ipv6_address = [i for i in ipv6_info if not i.startswith('fe80::')]
                return {"ipv4": ipv4_address, "ipv6": ipv6_address}
        except Exception as ex:
            logs.error(ex)
            return None

    @staticmethod
    def analysis_results(pkts_list, args: Namespace = None, call_name=None) -> None:
        """
        解析结果并存入队列
        :param args:
        :param pkts_list:
        :param DHCPv6:
        :param filter:
        :return:
        """
        filter_ip = args.dhcp_server
        call_func_name = getmodule(stack()[1][0])
        call_mod = call_func_name.__name__

        for pkt in pkts_list:
            if 'dhcp4' in call_mod:
                Tools.analysis_results_v4(pkt, args, filter_ip)
            else:
                Tools.analysis_results_v6(pkt, args, filter_ip, call_name)

    @staticmethod
    def analysis_results_v4(pkt, args, filter_ip) -> None:
        if pkt[IP].src == filter_ip:
            if pkt[DHCP].options[0][1] == 2:
                pkt_result.get('dhcp4_offer').put(pkt)
                Tools.print_formart(pkt, args.debug)
            elif pkt[DHCP].options[0][1] == 5:
                pkt_result.get('dhcp4_ack').put(pkt)
                Tools.print_formart(pkt, args.debug)
            elif pkt[DHCP].options[0][1] == 6:
                pkt_result.get('dhcp4_nak').put(pkt)
                Tools.print_formart(pkt, args.debug)
        else:
            logs.info('没有监听到 server 应答报文！,请检查是否有多个DHCP server影响监听结果')

    @staticmethod
    def analysis_results_v6(pkt, args, filter_ip, call_name=None, DHCPv6=None) -> None:
        if pkt[IPv6].src == filter_ip:
            if pkt[DHCPv6].msgtype == 2:
                try:
                    assert pkt[DHCP6OptIAAddress].addr
                    pkt_result.get('dhcp6_advertise').put(pkt)
                    if call_name is None:
                        Tools.print_formart(pkt, args.debug)
                except Exception as ex:
                    try:
                        assert pkt[DHCP6OptIAPrefix].prefix
                        pkt_result.get('dhcp6_advertise').put(pkt)
                        if call_name is None:
                            Tools.print_formart(pkt, args.debug)
                    except Exception as ex:
                        logs.error('应答报文中没有携带分配ip！')
                        assert False
            elif pkt[DHCPv6].msgtype == 7:
                pkt_result.get('dhcp6_reply').put(pkt)
                if call_name is None:
                    Tools.print_formart(pkt, args.debug)

            elif pkt[DHCPv6].msgtype == 13:
                ether_ipv6_udp = Ether() / IPv6(src=pkt[IPv6].src) / UDP()
                relay_pkt = ether_ipv6_udp / pkt[DHCP6OptRelayMsg].message
                Tools.analysis_results(pkts_list=relay_pkt, args=args, call_name=1)
                Tools.print_formart(pkt, args.debug)
        else:
            logs.info('没有监听到 server 应答报文！,请检查是否有多个DHCP server影响监听结果')

    @staticmethod
    def print_formart(pkt, debug) -> None:
        """
        格式化打印
        :param pkt:
        :param level:
        :return:
        """
        response_dict = {}
        if debug:
            pkt.show2()
        else:
            detail_info = pkt[UDP][1:].summary()
            mac = str2mac(global_var.get('generate_mac')) or ''
            if pkt.payload.name == 'IPv6':
                src_dst = pkt[IPv6].mysummary().split('(')[0]
                response_dict.update({"info": "{}".format(detail_info.split('/')[0])})
                content_format = Tools.print_formart_v6(pkt, response_dict, mac)
            else:
                src_dst = pkt[IP].mysummary().split('udp')[0]
                response_dict.update({"info": "{}".format(detail_info)})
                content_format = Tools.print_formart_v4(pkt, response_dict, mac)
            logs.info(content_format)
        Tools.record_pkt_num(pkt)

    @staticmethod
    def print_formart_v4(pkt, response_dict, mac) -> str:
        """
        DHCPv6格式化打印
        :param pkt:
        :param response_dict:
        :param mac:
        :return:
        """
        yiaddr = pkt[BOOTP].yiaddr
        response_dict.update({"yiaddr": yiaddr})
        yiaddr = response_dict.get('yiaddr') or ''
        info = response_dict.get('info') or ''
        content_format = "v4 | {:<} | {:<15} | {:<}".format(mac, yiaddr, info)
        return content_format

    @staticmethod
    def print_formart_v6(pkt, response_dict, mac) -> str:
        """
        DHCPv4格式化打印
        :param pkt:
        :param response_dict:
        :param mac:
        :return:
        """
        try:
            addr = pkt[DHCP6OptIAAddress].addr
            response_dict.update({"addr": addr})
            prefix = pkt[DHCP6OptIAPrefix].prefix
            response_dict.update({"prefix": prefix})
        except Exception as ex:
            if 'DHCP6OptIAAddress' in str(ex):
                try:
                    prefix = pkt[DHCP6OptIAPrefix].prefix
                    response_dict.update({"prefix": prefix})
                except:
                    pass
        addr = str(response_dict.get('addr') or '')
        prefix = str(response_dict.get('prefix') or '')
        info = str(response_dict.get('info') or '')
        content_format = "v6 | {:<} | NA: {:<15} | PD: {:<} | {:<}".format(mac, addr, prefix, info)
        return content_format

    @staticmethod
    def record_pkt_num(pkt, DHCPv6=None) -> None:
        try:
            for i in dhcp6types:
                if pkt[DHCPv6].msgtype == i:
                    summary_result[dhcp6types.get(i)] += 1
                    if pkt[DHCPv6].msgtype in (12, 13):
                        Tools.record_pkt_num(pkt[DHCP6OptRelayMsg].message)
        except:
            for v in DHCPTypes.values():
                pkt_type = pkt[DHCP].options[0][1]
                if isinstance(pkt_type, int):
                    pkt_type = DHCPTypes.get(pkt_type)
                if pkt_type == v:
                    summary_result[v.upper()] += 1

    @staticmethod
    def rate_print(text_tips, sleep_time) -> None:
        """
        倒计时打印
        :param text_tips:
        :param sleep_time:
        :return:
        """
        if sleep_time != 0:
            for i in range(sleep_time, 0, -1):
                if i == 1:
                    print("\r", text_tips, '倒计时', "{}".format(i), '', end="\n", flush=True)
                else:
                    print("\r", text_tips, '倒计时', "{}".format(i), '', end="", flush=True)
                time.sleep(1)

    @staticmethod
    def get_version_desc() -> str:
        version_desc = """
        dhcptool@1.1:
        2. 支持广播模式下配置指定的中继
        """
        return version_desc

    @staticmethod
    def check_ip_version(address) -> Optional[str]:
        # 尝试解析地址
        try:
            ip = ipaddress.ip_address(address)
            if isinstance(ip, ipaddress.IPv4Address):
                return 'ipv4'
            elif isinstance(ip, ipaddress.IPv6Address):
                return 'ipv6'
        except ValueError:
            print("无法解析该地址。")

    @staticmethod
    def read_yaml_config() -> dict:
        os_info = platform.system()
        yaml_config_path = 'dhcptool.yaml' if os_info == 'Windows' else '/etc/dhcptool.yaml'
        # 读取YAML文件
        with open(yaml_config_path, 'r', encoding='utf-8') as file:
            data = yaml.safe_load(file)
        return data

    @staticmethod
    def merge_args(cmd_args, ip_type) -> Namespace:
        """
        合并 config.yaml和  cmd_args的参数
        Args:
            cmd_args:
            ip_type:

        Returns: cmd_args

        """
        try:
            # 读取config.yaml中的参数配置
            config_args = Tools.read_yaml_config()
            # 根据ip_type 选择 ipv4的参数还是ipv6的参数
            config_args = config_args['v4'] if ip_type == 'ipv4' else config_args['v6']
            # 开始合并 options选项参数
            config_options = config_args.get('options') or {}
            if cmd_args.options:
                # 如果config_options存在参数
                cmd_options = [i.split('=') for i in cmd_args.options.split('&')]
                cmd_options = dict(cmd_options)
                # 将cmd_options参数覆盖config_option参数
                config_options.update(cmd_options)
            if config_options:
                config_options = {str(key): value for key, value in config_options.items()}
                # 将config_options转化为cmd_args格式的options结果
                config_options = list(config_options.items())
                config_options = ['='.join(option) for option in config_options]
                config_options = '&'.join(config_options)
            # 合并常规参数
            if cmd_args.ip_src is None and config_args.get('ip_src'):
                cmd_args.ip_src = config_args.get('ip_src')
            if cmd_args.dhcp_server is None and config_args.get('dhcp_server'):
                cmd_args.dhcp_server = config_args.get('dhcp_server')
            if cmd_args.relay_forward is None and config_args.get('relay_forward'):
                cmd_args.relay_forward = config_args.get('relay_forward')
            cmd_args.options = config_options
        except Exception as ex:
            pass
        return cmd_args

if __name__ == '__main__':
    ipv4, ipv6 = Tools.get_local_address()
    print(ipv4, ipv6)



# -*- coding: utf-8 -*-
# @Time    : 2022/10/23 18:59
# @Author  : mf.liang
# @File    : env_args.py
# @Software: PyCharm
# @desc    :

import queue
import random
from dhcptool.Logings import Logings

# 事务id(可选)
xid = random.randint(1, 900000000)
summary_result = {}
global_var = {"tag": 0}
pkt_result = {
    "dhcp6_advertise": queue.Queue(),
    "dhcp6_reply": queue.Queue(),
    "dhcp4_offer": queue.Queue(),
    "dhcp4_ack": queue.Queue(),
    "dhcp4_nak": queue.Queue(),
}
logs = Logings("DHCP")



# -*- coding: utf-8 -*-
# @Time    : 2022/10/23 19:00
# @Author  : mf.liang
# @File    : dhcp_pkt_v6.py
# @Software: PyCharm
# @desc    :
import threading
from time import sleep
from scapy.layers.dhcp import BOOTP
from scapy.layers.dhcp6 import DHCP6, All_DHCP_Relay_Agents_and_Servers
from scapy.layers.inet import UDP, IP
from scapy.layers.inet6 import IPv6
from scapy.layers.l2 import Ether
from scapy.plist import PacketList
from scapy.sendrecv import sendp, AsyncSniffer
from dhcptool.tools import Tools
from typing import Optional


class BasePkt:

    def __init__(self, args) -> None:
        self.args = args
        self.timeout = 200 / 1000
        self.mac = Tools.get_mac(self.args)
        self.xid = Tools.get_xid_by_mac(self.mac)
        self.ether = Ether()
        self.ip = IP(src='0.0.0.0', dst='255.255.255.255')
        self.ipv6 = IPv6(dst=All_DHCP_Relay_Agents_and_Servers)
        self.udp = UDP()
        self.bootp = BOOTP(xid=self.xid, chaddr=self.mac)
        self.dhcp6 = DHCP6(trid=self.xid)
        self.start_event = threading.Event()

    def async_sniff(self, async_sniff_args, pkt):
        Tools.print_formart(pkt, self.args.debug)
        async_sniff_result = AsyncSniffer(
            iface=self.args.iface if self.args.iface else None,
            started_callback=lambda: self.start_event.set(),
            **async_sniff_args
        )
        # 开始监听
        async_sniff_result.start()
        self.start_event.wait()
        sleep(0.01)
        if self.args.iface:
            for iface in self.args.iface:
                sendp(pkt, verbose=0, iface=iface)
        else:
            sendp(pkt, verbose=0)
        async_sniff_result.join()
        # 获取结果
        response_pkts = async_sniff_result.results
        return response_pkts

    def send_dhcp6_pkt(self, pkt) -> Optional[PacketList]:
        """
        发送并接收 dhcp6 数据包
        :param pkt:
        :return:
        """
        async_sniff_args = {"filter": f'port 547 and src host {self.args.dhcp_server}', "count": 1, "timeout": self.timeout}
        return self.async_sniff(async_sniff_args, pkt)

    def send_dhcp4_pkt(self, pkt) -> Optional[PacketList]:
        """
        发送并接收 dhcp4 数据包
        :param pkt:
        :return:
        """
        async_sniff_args = {"filter": f'port 67 and src host {self.args.dhcp_server}', "count": 1, "timeout": self.timeout}
        response = self.async_sniff(async_sniff_args, pkt)
        return response



# -*- coding: utf-8 -*-
# @Time    : 2022/10/23 19:06
# @Author  : mf.liang
# @File    : dhcp6_controller.py
# @Software: PyCharm
# @desc    :
from queue import Empty
from scapy.layers.dhcp6 import dhcp6types
from dhcptool.dhcp_pkt_v6 import Dhcp6Pkt
from dhcptool.env_args import summary_result, logs, pkt_result, global_var
from dhcptool.tools import Tools


class Dhcp6Controller(Dhcp6Pkt):

    def __init__(self, args) -> None:
        super(Dhcp6Controller, self).__init__(args)
        self.args = args

    def run(self) -> None:
        """
        执行 发包测试入口
        :return:
        """
        for i in dhcp6types.values():
            summary_result[i] = 0
        for i in range(int(self.args.num)):
            global_var['tag'] = i
            try:
                send_pkts = {
                    self.args.renew: self.send_renew if self.args.single else self.send_solicit_advertise_request_reply_renew,
                    self.args.release: self.send_release if self.args.single else self.send_solicit_advertise_request_reply_release,
                    self.args.decline: self.send_decline if self.args.single else self.send_solicit_advertise_request_reply_decline,
                    self.args.solicit: self.send_solicit,
                }
                send_pkts.get(True, self.send_solicit_advertise_request_reply)()
            except Empty as ex:
                logs.info('没有接收到应答报文！')
            except AssertionError as ex:
                logs.info('应答报文未包含分配ip！')
            print('-' * 60)
            pkt_result.get('dhcp6_reply').queue.clear()
        str_summary_result = str(summary_result).replace('{', '').replace('}', '').replace("'", '')
        print(str_summary_result)

    def send_solicit_advertise_request_reply(self) -> None:
        """
        发送  dhcp6 完整分配流程
        :return:
        """
        self.__init__(self.args)
        solicit_pkt = self.dhcp6_solicit()
        res = self.send_dhcp6_pkt(solicit_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)
        request_pkt = self.dhcp6_request()
        res = self.send_dhcp6_pkt(request_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)
        Tools.rate_print('sleep time', self.args.sleep_time)

    def send_solicit_advertise_request_reply_renew(self) -> None:
        """
        分配完地址后进行 更新租约
        :return:
        """
        self.send_solicit_advertise_request_reply()
        renew_pkt = self.dhcp6_renew()
        res = self.send_dhcp6_pkt(renew_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_solicit_advertise_request_reply_release(self) -> None:
        """
        分配完地址后进行 释放地址
        :return:
        """
        self.send_solicit_advertise_request_reply()
        release_pkt = self.dhcp6_release()
        res = self.send_dhcp6_pkt(release_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_solicit_advertise_request_reply_decline(self) -> None:
        """
        分配完地址后进行 释放地址
        :return:
        """
        self.send_solicit_advertise_request_reply()
        decline_pkt = self.dhcp6_decline()
        res = self.send_dhcp6_pkt(decline_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_solicit(self) -> None:
        """
        单独的solicit包
        :return:
        """
        discover_pkt = self.dhcp6_solicit()
        res = self.send_dhcp6_pkt(discover_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_renew(self) -> None:
        """
        单独的renew包
        :return:
        """
        renew_pkt = self.dhcp6_renew()
        res = self.send_dhcp6_pkt(renew_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_decline(self) -> None:
        """
        单独的decline包
        :return:
        """
        decline_pkt = self.dhcp6_decline()
        res = self.send_dhcp6_pkt(decline_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_release(self) -> None:
        """
        单独的release包
        :return:
        """
        release_pkt = self.dhcp6_release()
        res = self.send_dhcp6_pkt(release_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)




# -*- coding: utf-8 -*-
# @Time    : 2022/10/25 0:13
# @Author  : mf.liang
# @File    : Logings.py
# @Software: PyCharm
# @desc    :

import sys
import logging


class Logings(logging.Logger):
    def __init__(self, name, level=logging.INFO, file=None):
        """
        :param name: 日志名字
        :param level: 级别
        :param file: 日志文件名称
        """
        super().__init__(name, level)
        # 设置日志格式
        formatter = logging.Formatter("%(asctime)s,%(msecs)03d | %(message)s", datefmt="%H:%M:%S")
        # 文件输出渠道
        if file:
            handle2 = logging.FileHandler(file, encoding="utf-8")
            handle2.setFormatter(formatter)
            self.addHandler(handle2)
        # 控制台渠道
        handle1 = logging.StreamHandler(sys.stdout)
        handle1.setFormatter(formatter)
        self.addHandler(handle1)



# -*- coding: utf-8 -*-
# @Time    : 2022/10/23 16:33
# @Author  : mf.liang
# @File    : main.py
# @Software: PyCharm
# @desc    : dhcp 命令行接受，本工具为循环发包，未考虑实现并发
import argparse
import time

from dhcptool.dhcp4_controller import Dhcp4Controller
from dhcptool.dhcp6_controller import Dhcp6Controller
from dhcptool.env_args import logs
from dhcptool.extend_tools_controller import ExtendToolsController
from dhcptool.tools import Tools


def parse_cmd_args_common(subparsers) -> None:
    """
    DHCPv4和DHCPv6公共解析参数
    :param subparsers: 
    :return: 
    """
    # Boolean选项
    subparsers.add_argument("--debug", "-debug", action='store_true', help='debug模式，能查看更详细的报文内容')
    subparsers.add_argument("--single", "-single", action='store_true',
                            help='dhcptool [v4|v6] -s [ipv4|ipv6] -mt inform -single -o 50=[ipv4]')
    subparsers.add_argument("--renew", "-renew", action='store_true', help='dhcptool [v4|v6] -f [ipv4|ipv6] -renew')
    subparsers.add_argument("--release", "-release", action='store_true', help='发起release请求')
    subparsers.add_argument("--decline", "-decline", action='store_true', help='发起decline请求')
    subparsers.add_argument("--fixe", "-fixe", action='store_true', help="固定mac值进行发包")
    subparsers.add_argument("--broadcast", "-b", action='store_true', help="广播发包")
    # 接收参数选项
    subparsers.add_argument("--ip_src", "-src", default=None, help='dhcptool [v4|v6] -s [ipv4|ipv6] -src [ipv4|ipv6]')
    subparsers.add_argument("--dhcp_server", "-s", help="dhcptool [v4|v6] -s [ipv4|ipv6]")
    subparsers.add_argument('--iface', "-i", nargs='+', help='dhcptool [v4|v6] -i eth0 eth1')
    subparsers.add_argument("--relay_forward", "-rf", type=str, default=None,
                            help='dhcptool [v4|v6] -f [ipv4|ipv6] -rf [ipv4|ipv6]')
    subparsers.add_argument("--mac", "-mac", default=None, help='指定用户mac')
    subparsers.add_argument("--encoding", "-e", default='utf-8', help='为部分option的值指定编码')
    subparsers.add_argument("--num", "-n", type=int, default=1, help="发送报文数量")
    subparsers.add_argument("--sleep_time", "-st", type=int, default=0, help='发包过程中等待时间')
    subparsers.add_argument("--options", "-o", default=None,
                            help='dhcptool [v4|v6] -f [ipv4|ipv6] -o [code]=[value]&[code]=[value] [dhcptool v4 -s '
                                 '192.168.31.134 -o [16=1f3……&14=''][18="01|eth 2/1/4:114.12 ZTEOLT001/1/1/5/0/1/000000000000001111111152,02|bf:e5:34:12:39:04,06|mf.liang"][60=60:000023493534453……][6=12,'
                                 '7][50=192.168.31.199]')


def parse_cmd_args_dhcp4() -> None:
    """
    # 解析dhcp4参数
    :return:
    """
    parse_cmd_args_common(subparsers_4)
    subparsers_4.add_argument("--discover", "-discover", action='store_true', help='发起discover请求')
    subparsers_4.add_argument("--request", "-request", action='store_true', help='发起request请求')
    subparsers_4.add_argument("--inform", "-inform", action='store_true', help='发起inform请求')
    subparsers_4.add_argument("--nak", "-nak", action='store_true', help='模拟异常报文让服务器应答nak响应报文')


def parse_cmd_args_dhcp6() -> None:
    """
    解析dhcp6参数
    :return:
    """
    parse_cmd_args_common(subparsers_6)
    subparsers_6.add_argument("--na", "-na", action='store_true', help='')
    subparsers_6.add_argument("--ta", "-ta", action='store_true', help='')
    subparsers_6.add_argument("--pd", "-pd", action='store_true', help='')
    subparsers_6.add_argument("--duid", "-duid", default=None, help='duid')
    subparsers_6.add_argument("--solicit", "-solicit", action='store_true', help='发起solicit请求')


def parse_cmd_args_tools() -> None:
    """
    解析tools参数
    :return:
    """
    parse_cmd_args_common(subparsers_tools)
    subparsers_tools.add_argument("--generate_mac", "-gm", choices=["random", "asc", "desc"], help='dhcptool tools -gm ["random", "asc", "desc"]')
    subparsers_tools.add_argument("--generate_address", "-ga", help='dhcptool tools -ga')
    subparsers_tools.add_argument("--ipv6", "-v6", action='store_true', help='dhcptool tools -v6')
    subparsers_tools.add_argument("--ipv4", "-v4", action='store_true', help='dhcptool tools -v4')
    subparsers_tools.add_argument("--na", "-na", action='store_true', help='')
    subparsers_tools.add_argument("--pd", "-pd", type=int, help='dhcptool tools -pd 1000:0:0:31::0/64')


def exec_dhcp4(args) -> None:
    """
    DHCPv4发包
    :param args:
    :return:
    """
    start = time.perf_counter()
    merge_cmd_args = Tools.merge_args(args, 'ipv4')
    dhcp4_controller = Dhcp4Controller(merge_cmd_args)
    dhcp4_controller.run()
    end = time.perf_counter()
    print(f"执行耗时: {end - start:.6f} 秒")


def exec_dhcp6(args) -> None:
    """
    DHCPv6发包
    :param args:
    :return:
    """
    start = time.perf_counter()
    merge_cmd_args = Tools.merge_args(args, 'ipv6')
    dhcp6_controller = Dhcp6Controller(merge_cmd_args)
    dhcp6_controller.run()
    end = time.perf_counter()
    print(f"执行耗时: {end - start:.6f} 秒")


def exec_tools(args) -> None:
    """
    执行 tools中的功能
    :param args:
    :return:
    """
    extend_tool_controller = ExtendToolsController(args)
    if args.generate_address:
        extend_tool_controller.calculate_address_range()
    if args.generate_mac:
        extend_tool_controller.generate_print_mac()


epilog="""
示例：
dhcptool v6 -s 1000:0:0:21::18 -src 1000:0:0:21::22 -rf 1000:0:0:21::22 -duid dca52e9895d81212
"""

parser = argparse.ArgumentParser(conflict_handler='resolve')


parser.add_argument("--version", "-v", help="查看dhcptool版本信息", action='version', version=Tools.get_version_desc())
subparsers = parser.add_subparsers(
    help='[s|f] [debug] [single] [renew|release|decline|inform|nak] [mac] [o] [n] [st] [np]')
subparsers_4 = subparsers.add_parser('v4', help="DHCPv4 发包帮助信息")
subparsers_4.set_defaults(func=exec_dhcp4)
subparsers_6 = subparsers.add_parser('v6', help='DHCPv6 发包帮助信息')
subparsers_6.set_defaults(func=exec_dhcp6)
subparsers_tools = subparsers.add_parser('tools', help='扩展工具')
subparsers_tools.set_defaults(func=exec_tools)
parse_cmd_args_dhcp4()
parse_cmd_args_dhcp6()
parse_cmd_args_tools()


def dhcp_main() -> None:
    """
    dhcp执行函数入口
    :return:  v6 -s 1000:0:0:31::11 -n 5
    """
    cmd_args = parser.parse_args()
    args_dict = vars(cmd_args)
    if args_dict:
        str_args_dict = str(args_dict).replace('{', '').replace('}', '').replace("'", '')
        if cmd_args.debug:
            logs.info(f"args: {str_args_dict}")
        # 开启执行
        cmd_args.func(cmd_args)
    else:
        while True:
            argv_cmd = input('dhcptool:\t')
            argv_list = argv_cmd.split(' ')
            args = parser.parse_args(argv_list)
            # 开启执行
            args.func(args)

def dhcp_code_call_main(args_cmd_list) -> None:
    """通过代码调用执行dhcptool

    Args:
        args_cmd_list (_type_): 命令解析参数
    """
    args = parser.parse_args(args_cmd_list)
    args_dict = vars(args)
    str_args_dict = str(args_dict).replace('{', '').replace('}', '').replace("'", '')
    if args.debug:
        logs.info(f"args: {str_args_dict}")
    args.func(args)


if __name__ == '__main__':

    dhcp_main()




# -*- coding: utf-8 -*-
# @Time    : 2022/10/23 19:06
# @Author  : mf.liang
# @File    : dhcp4_controller.py
# @Software: PyCharm
# @desc    :
from queue import Empty
from scapy.layers.dhcp import DHCPTypes
from dhcptool.dhcp_pkt_v4 import Dhcp4Pkt
from dhcptool.env_args import summary_result, logs, pkt_result, global_var
from dhcptool.tools import Tools


class Dhcp4Controller(Dhcp4Pkt):

    def __init__(self, args) -> None:
        super(Dhcp4Controller, self).__init__(args)
        self.args = args

    def run(self) -> None:
        """
        执行 发包测试入口
        :return:
        """
        for i in DHCPTypes.values():

            summary_result[i.upper()] = 0
        for i in range(int(self.args.num)):
            global_var['tag'] = i
            try:
                send_pkts = {
                    self.args.renew: self.send_request if self.args.single else self.send_discover_offer_request_ack_renew,
                    self.args.release: self.send_release if self.args.single else self.send_discover_offer_request_ack_release,
                    self.args.inform: self.send_inform if self.args.single else self.send_discover_offer_request_ack_inform,
                    self.args.request: self.send_request if self.args.single else None,
                    self.args.discover: self.send_discover,
                    self.args.nak: self.send_discover_offer_request_nak,
                    self.args.decline: self.send_decline if self.args.single else self.send_discover_offer_request_ack_decline,
                }
                send_pkts.get(True, self.send_discover_offer_request_ack)()
            except Empty as ex:
                logs.info('没有接收到应答报文！')
            except AssertionError as ex:
                logs.info('应答报文未包含分配ip！')
            except Exception as ex:
                logs.info(f"warn: {ex}")
            print('-' * 60)
            pkt_result.get('dhcp4_ack').queue.clear()
        str_summary_result = str(summary_result).replace('{', '').replace('}', '').replace("'", '')
        print(str_summary_result)

    def send_discover_offer_request_ack(self):
        """
        发送  dhcp4 完整分配流程
        :return:
        """
        self.__init__(self.args)
        discover_pkt = self.dhcp4_discover()
        res = self.send_dhcp4_pkt(discover_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)
        request_pkt = self.dhcp4_request()
        ack_pkt = self.send_dhcp4_pkt(request_pkt)
        Tools.analysis_results(pkts_list=ack_pkt, args=self.args)
        Tools.rate_print('sleep time', self.args.sleep_time)
        return request_pkt

    def send_discover_offer_request_nak(self) -> None:
        """
        发送  dhcp4 完整分配流程
        :return:
        """
        self.__init__(self.args)
        discover_pkt = self.dhcp4_discover()
        res = self.send_dhcp4_pkt(discover_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)
        request_pkt = self.dhcp4_exception_request()
        ack_pkt = self.send_dhcp4_pkt(request_pkt)
        Tools.analysis_results(pkts_list=ack_pkt, args=self.args)
        

    def send_discover_offer_request_ack_renew(self) -> None:
        """
        发起 更新租约 请求
        :return:
        """
        request_pkt = self.send_discover_offer_request_ack()
        ack_pkt = self.send_dhcp4_pkt(request_pkt)
        Tools.analysis_results(pkts_list=ack_pkt, args=self.args)

    def send_discover_offer_request_ack_decline(self) -> None:
        """
        发起 冲突租约 请求
        :return:
        """
        self.send_discover_offer_request_ack()
        decline_pkt = self.dhcp4_decline()
        self.send_dhcp4_pkt(decline_pkt)

    def send_discover_offer_request_ack_release(self) -> None:
        """
        发起 释放租约 请求
        :return:
        """
        self.send_discover_offer_request_ack()
        release_pkt = self.dhcp4_release()
        self.send_dhcp4_pkt(release_pkt)

    def send_discover_offer_request_ack_inform(self) -> None:
        """
        发起 inform 请求
        :return:
        """
        self.send_discover_offer_request_ack()
        inform_pkt = self.dhcp4_inform()
        ack_pkt = self.send_dhcp4_pkt(inform_pkt)
        Tools.analysis_results(pkts_list=ack_pkt, args=self.args)

    def send_discover(self) -> None:
        """
        单独的discover包
        :return:
        """
        discover_pkt = self.dhcp4_discover()
        res = self.send_dhcp4_pkt(discover_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_request(self) -> None:
        """
        单独的request包
        :return:
        """
        request_pkt = self.dhcp4_request()
        res = self.send_dhcp4_pkt(request_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_inform(self) -> None:
        """
        单独的inform包
        :return:
        """
        inform_pkt = self.dhcp4_inform()
        res = self.send_dhcp4_pkt(inform_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_decline(self) -> None:
        """
        单独的decline包
        :return:
        """
        decline_pkt = self.dhcp4_decline()
        res = self.send_dhcp4_pkt(decline_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)

    def send_release(self) -> None:
        """
        单独的release包
        :return:
        """
        release_pkt = self.dhcp4_release()
        res = self.send_dhcp4_pkt(release_pkt)
        Tools.analysis_results(pkts_list=res, args=self.args)



# -*- coding: utf-8 -*-
# @Time    : 2022/10/23 19:33
# @Author  : mf.liang
# @File    : options.py
# @Software: PyCharm
# @desc    :
import binascii
from functools import reduce
from typing import Text, Optional
from scapy.layers.dhcp6 import DHCP6OptVendorClass, VENDOR_CLASS_DATA, DHCP6OptIfaceId, DHCP6OptStatusCode, \
    DHCP6OptRapidCommit, DHCP6OptOptReq, \
    DHCP6OptIAAddress, DHCP6OptIAPrefix, DHCP6OptClientFQDN, DHCP6OptSubscriberID
from scapy.layers.inet import IP
from dhcptool.env_args import logs


class Options:
    def __init__(self, args):
        self.args = args

    def parse_options(self):
        options_list = self.args.options
        if options_list:
            options_list = [i.split('=') for i in options_list.split('&')]
            return options_list


class Dhcp4Options(Options):

    def __init__(self, args):
        self.args = args
        super(Dhcp4Options, self).__init__(args=self.args)

    def make_options_list(self) -> list:
        """
        制作 options
        :return:
        """
        options = []
        options_list = self.parse_options()
        if options_list is not None:
            for index, i in enumerate(options_list):
                if int(i[0]) == 12:
                    options.append(self.option_12(hostname=i[1]))
                if int(i[0]) == 7:
                    options.append(self.option_7(log_server=i[1]))
                if int(i[0]) == 60:
                    options.append(self.option_60(vendor_class_id=i[1]))
                if int(i[0]) == 82:
                    options.append(self.option_82(i[1]))
                if int(i[0]) == 55:
                    options.append(self.option_55(param_req_list=i[1]))
                if int(i[0]) == 50:
                    options.append(self.option_50(requested_addr=i[1]))
                if int(i[0]) == 51:
                    options.append(self.option_51(lease_time=i[1]))
                if int(i[0]) == 54:
                    options.append(self.option_54(server_id=i[1]))
                if int(i[0]) == 2:
                    options.append(self.option_2(time_zone=i[1]))
                if int(i[0]) == 3:
                    options.append(self.option_3(router=i[1]))
                if int(i[0]) == 13:
                    options.append(self.option_13(boot_size=i[1]))
                if int(i[0]) == 15:
                    options.append(self.option_15(domain=i[1]))
                if int(i[0]) == 19:
                    options.append(self.option_19(ip_forwarding=i[1]))
                if int(i[0]) == 23:
                    options.append(self.option_23(default_ttl=i[1]))
                if int(i[0]) == 61:
                    options.append(self.option_61(client_id=i[1]))
                if int(i[0]) == 97:
                    options.append(self.option_97(pxe_client_machine_identifier=i[1]))
                if int(i[0]) == 136:
                    options.append(self.option_136(pana_agent=i[1]))
                if int(i[0]) == 141:
                    options.append(self.option_141(sip_ua_service_domains=i[1]))
                if int(i[0]) == 161:
                    options.append(self.option_161(mud_url=i[1]))
        options.append('end')
        return options

    def option_12(self, hostname='') -> tuple:
        hostname = hostname.encode(self.args.encoding)
        return 'hostname', hostname

    def option_7(self, log_server='0.0.0.0') -> tuple:
        return 'log_server', log_server

    def option_60(self, vendor_class_id='') -> tuple:
        """
        拼接option60的函数
        :param vendor_class_id:
        :return:
        ./dhcptool v4 -s 192.168.31.134 -o 60=$(radtools passwd mf@liang admin123)
        """
        try:
            hex = vendor_class_id.encode("utf-8")
            vendor_class_id = binascii.unhexlify(hex)
            return 'vendor_class_id', vendor_class_id
        except:
            return 'vendor_class_id', vendor_class_id

    def option_82(self, value='') -> tuple:
        """
        TODO: option82 和别的option同时使用时，sub_option存在顺序错乱的问题
        ./dhcptool v4 -s 192.168.31.116 -o "60=$(radtools passwd user1@itv.com test123)&82="01|eth 2/1/4:114.12 ZTEOLT001/1/1/5/0/1/000000000000001111111152,02|bf:e5:34:12:39:04,06|mf.liang"&12=yamu.com&7=2.2.2.2"
        :param value:
        :return:
        "01|eth 2/1/4:114.12 ZTEOLT001/1/1/5/0/1/000000000000001111111152,02|bf:e5:34:12:39:04,06|mf.liang"
        ,号分割suboption
        |号前面是suboption id， 后面是值
        """
        suboption_cmd_list = value.split(',')
        suboption_list = []
        for i in suboption_cmd_list:
            suboption = i.split('|')
            try:
                #计算长度
                value_len = hex(len(suboption[1]))[2:]
                #计算hex
                hex_value = suboption[1].encode(self.args.encoding).hex()
                #组合
                suboption_txt = str(suboption[0]) + str(value_len) + hex_value
                hex_value = suboption_txt.encode(self.args.encoding)
                value = binascii.unhexlify(hex_value)
            except:
                hex_value = suboption[1].encode(self.args.encoding)
                value = binascii.unhexlify(hex_value)
            suboption_list.append(value)
        suboption_bytes = b''.join(suboption_list)
        return 'relay_agent_information', suboption_bytes

    def option_55(self, param_req_list='') -> tuple:
        param_req_list = [int(i) for i in param_req_list.split(',')]
        return 'param_req_list', param_req_list

    def option_50(self, requested_addr='192.168.0.1') -> tuple:
        return 'requested_addr', requested_addr

    def option_51(self, lease_time: int = 43200) -> tuple:
        """

        :param lease_time: 租约时间
        :return:
        """
        return 'lease_time', int(lease_time)

    def option_54(self, server_id='0.0.0.0') -> tuple:
        return 'server_id', server_id

    def option_61(self, client_id: str) -> tuple:
        """

        :param client_id: client_id
        :return:
        """
        return 'client_id', client_id

    def option_2(self, time_zone: int = 500) -> tuple:
        """
        :param time_zone: 时区
        :return:
        """
        return 'time_zone', int(time_zone)

    def option_3(self, router='0.0.0.0') -> tuple:
        """

        :param router: 路由
        :return:
        """
        return 'router', router

    def option_13(self, boot_size=1000) -> tuple:
        """

        :param boot_size:
        :return:
        """
        return 'boot-size', int(boot_size)

    def option_15(self, domain: Text) -> tuple:
        """

        :param domain: 域名
        :return:
        """
        return 'domain', domain

    def option_19(self, ip_forwarding: bool) -> tuple:
        """

        :param ip_forwarding: ip转发
        :return:
        """
        return 'ip_forwarding', bool(ip_forwarding)

    def option_23(self, default_ttl: int) -> tuple:
        """

        :param default_ttl: 默认ip ttl值
        :return:
        """
        return 'default_ttl', default_ttl

    def option_97(self, pxe_client_machine_identifier: str) -> tuple:
        """
        pxe客户端机器标识符
        Args:
            pxe_client_machine_identifier:

        Returns:

        """
        return 'pxe_client_machine_identifier', pxe_client_machine_identifier

    def option_136(self, pana_agent: IP) -> tuple:
        """

        Args:
            pana_agent:

        Returns:

        """
        return 'pana-agent', pana_agent

    def option_141(self, sip_ua_service_domains: str) -> tuple:
        return 'sip_ua_service_domains', sip_ua_service_domains

    def option_161(self, mud_url: str) -> tuple:
        return 'mud-url', mud_url


class Dhcp6Options(Options):

    def __init__(self, args):
        self.args = args
        super(Dhcp6Options, self).__init__(args=self.args)

    def make_options_list(self) -> tuple:
        """
        制作 options
        :return:
        """
        options = DHCP6OptStatusCode()
        parent_options = []
        options_list = self.parse_options()
        if options_list is not None:
            for i in options_list:
                if int(i[0]) == 16:
                    options = self.option_16(i[1]) / options
                if int(i[0]) == 18:
                    if self.args.dhcp_server:
                        parent_options.append(self.option_18(i[1]))
                if int(i[0]) == 38:
                    parent_options.append(self.option_38(i[1]))
                if int(i[0]) == 6:
                    options = self.option_6(i[1]) / options
                if int(i[0]) == 14:
                    options = self.option_14() / options
                if int(i[0]) == 39:
                    options = self.option_39(i[1]) / options

        if parent_options:
            parent_options = reduce(lambda x, y: y / x, parent_options)
        return options, parent_options

    def option_16(self, account_pwd_hex: str) -> Optional[DHCP6OptVendorClass]:
        """
        python3 dhcptool.py v6 -s 1000::31:332b:d5ab:4457:fb60 -debug on -o "16=1f31014d65822107fcfd52000000006358c1cc2f31c57f7dd8b43d27edc570aba8e999ed46b5176fb38bb7a407d97010eeebba"
        :return:
        """
        try:
            if "0000" in account_pwd_hex[:5]:
                account_pwd_hex = account_pwd_hex[4:]
            vendor_class_data = VENDOR_CLASS_DATA(data=bytes.fromhex(account_pwd_hex))
            option16_pkt = DHCP6OptVendorClass(vcdata=vendor_class_data)
        except Exception as ex:
            logs.error(ex)
            return None
        return option16_pkt

    def option_18(self, ipoe_value: str) -> DHCP6OptIfaceId:
        """
        suxx@suxx:      eth 2/1/4:80.90 ZTEOLT001/1/1/5/0/1/
        python3 dhcptool.py v6 -s 1000::31:332b:d5ab:4457:fb60 -o "16=1f31014d65822107fcfd52000000006358c1cc2f31c57f7dd8b43d27edc570aba8e999ed46b5176fb38bb7a407d97010eeebba&18=eth 2/1/4:80.90 ZTEOLT001/1/1/5/0/1"
        :return:
        """
        option18_pkt = DHCP6OptIfaceId(ifaceid=ipoe_value)
        return option18_pkt

    def option_38(self, value: str) -> DHCP6OptSubscriberID:
        """
        suxx@suxx:      eth 2/1/4:80.90 ZTEOLT001/1/1/5/0/1/
        python3 dhcptool.py v6 -s 1000::31:332b:d5ab:4457:fb60 -o "16=1f31014d65822107fcfd52000000006358c1cc2f31c57f7dd8b43d27edc570aba8e999ed46b5176fb38bb7a407d97010eeebba&18=eth 2/1/4:80.90 ZTEOLT001/1/1/5/0/1"
        :return:
        """
        value = value.encode(self.args.encoding)
        option38_pkt = DHCP6OptSubscriberID(subscriberid=value)
        return option38_pkt

    def option_6(self, value) -> DHCP6OptOptReq:
        """
        Option Request
        :return:
        """
        if value:
            value_list = [int(i) for i in value.split(',')]
            option6_pkt = DHCP6OptOptReq(reqopts=value_list)
        else:
            option6_pkt = DHCP6OptOptReq()
        return option6_pkt

    def option_14(self) -> DHCP6OptRapidCommit:
        """
        Rapid Commit
        :return:
        """
        option14_pkt = DHCP6OptRapidCommit()
        return option14_pkt

    def sub_option_5(self, addr) -> DHCP6OptIAAddress:
        """
        DHCP6OptIAAddress   二级option,不能与 一级进行 option拼接
        :return:
        """
        option5_pkt = DHCP6OptIAAddress(addr=addr)
        return option5_pkt

    def sub_option_26(self, prefix) -> DHCP6OptIAPrefix:
        """
        DHCP6OptIAPrefix  二级option,不能与 一级进行 option拼接
        :return:
        """
        prefix, prefix_len = prefix.split('/')[0], prefix.split('/')[1]
        option26_pkt = DHCP6OptIAPrefix(prefix=prefix, plen=int(prefix_len))
        return option26_pkt

    def option_39(self, fqdn) -> DHCP6OptClientFQDN:
        """
        DHCP6OptClientFQDN
        :return:
        """
        fqdn = fqdn.encode(self.args.encoding)
        option39_pkt = DHCP6OptClientFQDN(flags=0, fqdn=fqdn)
        return option39_pkt



# coding = 'utf-8'
"""
@File:          extend_tools_controller.py
@Time:          2023/7/20 16:30
@Author:        mf.liang
@Email:         mf.liang@yamu.com
@Desc:          请注明模块要实现的功能

"""
import ipaddress
from scapy.volatile import RandMAC
from dhcptool.env_args import logs
from dhcptool.tools import Tools


class ExtendToolsController:

    def __init__(self, args):
        self.args = args

    def generate_print_mac(self) -> None:
        """
        生成并打印mac
        :return:
        """
        init_mac = str(RandMAC())
        for i in range(self.args.num if self.args.num != 1 else 100):
            if self.args.generate_mac == 'random':
                mac = str(RandMAC())
            elif self.args.generate_mac == 'asc':
                mac = Tools.mac_self_incrementing_or_subtracting(init_mac, i, 'ASC')
            else:
                mac = Tools.mac_self_incrementing_or_subtracting(init_mac, i, 'DESC')
            print(mac)

    def calculate_address_range(self) -> None:
        """
        计算IPv4/IPv6的地址范围
        :return:
        """
        address, mask_length = tuple(self.args.generate_address.split('/'))
        if self.args.ipv4:
            network = ipaddress.IPv4Network(address + '/' + str(mask_length), strict=False)
        if self.args.ipv6:
            network = ipaddress.IPv6Network(address + '/' + str(mask_length), strict=False)
            if self.args.pd:
                network = ipaddress.IPv6Network(address + '/' + str(mask_length), strict=False)
                network_pd = network.subnets(new_prefix=self.args.pd)
                pd_address = [str(ip) for ip in network_pd]
                logs.info(f"前缀地址：{pd_address}")

        logs.info(f"地址划分：{self.args.generate_address}")
        logs.info(f"地址类型：{network.version}")
        logs.info(f"主机掩码：{network.with_hostmask}")
        logs.info(f"地址数量：{network.num_addresses}")
        logs.info(f"起始范围：{network.network_address} - {network.broadcast_address}")
        addresses = [str(ip) for ip in network]
        logs.info(f"起始地址：{addresses[:256 // 2] + ['...'] + addresses[-256 // 2:]}")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@File    :   code_call_exec.py
@Time    :   2023/08/10 14:25:02
@Author  :   mf.liang
@Version :   1.0
@Contact :   mf.liang@outlook.com
@Desc    :
"""

from ipaddress import IPv4Address, IPv6Address
from typing import NewType, Text

from main import dhcp_code_call_main

MACAddress = NewType('MACAddress', str)


def dhcp4_tool(server: IPv4Address = None,
               relay: IPv4Address = None,
               filter: IPv4Address = None,
               ip_src: IPv4Address = None,
               single: bool = False,
               discover: bool = False,
               request: bool = False,
               inform: bool = False,
               nak: bool = False,
               renew: bool = False,
               release: bool = False,
               decline: bool = False,
               mac: MACAddress = None,
               num: int = 1,
               debug: bool = False,
               options: Text = None) -> None:
    """执行dhcp4发包

    Args:
        server (IPv4Address, optional): _description_. Defaults to None.
        relay (IPv4Address, optional): _description_. Defaults to None.
        filter (IPv4Address, optional): _description_. Defaults to None.
        ip_src (IPv4Address, optional): _description_. Defaults to None.
        single (bool, optional): _description_. Defaults to False.
        discover (bool, optional): _description_. Defaults to False.
        request (bool, optional): _description_. Defaults to False.
        inform (bool, optional): _description_. Defaults to False.
        nak (bool, optional): _description_. Defaults to False.
        renew (bool, optional): _description_. Defaults to False.
        release (bool, optional): _description_. Defaults to False.
        decline (bool, optional): _description_. Defaults to False.
        mac (MACAddress, optional): _description_. Defaults to None.
        num (int, optional): _description_. Defaults to 1.
        debug (bool, optional): _description_. Defaults to False.
        options (Text, optional): _description_. Defaults to None.
    """
    args_cmd_list = ['v4']

    dhcp_code_call_main(args_cmd_list)


def dhcp6_tool(server: IPv6Address = None,
               relay: IPv6Address = None,
               filter: IPv6Address = None,
               ip_src: IPv6Address = None,
               single: bool = False,
               na: bool = False,
               pd: bool = False,
               solicit: bool = False,
               renew: bool = False,
               release: bool = False,
               decline: bool = False,
               mac: MACAddress = None,
               num: int = 1,
               debug: bool = False,
               options: Text = None) -> None:
    """执行dhcp6发包

    Args:
        server (IPv6Address, optional): _description_. Defaults to None.
        relay (IPv6Address, optional): _description_. Defaults to None.
        filter (IPv6Address, optional): _description_. Defaults to None.
        ip_src (IPv6Address, optional): _description_. Defaults to None.
        single (bool, optional): _description_. Defaults to False.
        discover (bool, optional): _description_. Defaults to False.
        request (bool, optional): _description_. Defaults to False.
        inform (bool, optional): _description_. Defaults to False.
        nak (bool, optional): _description_. Defaults to False.
        renew (bool, optional): _description_. Defaults to False.
        release (bool, optional): _description_. Defaults to False.
        decline (bool, optional): _description_. Defaults to False.
        mac (MACAddress, optional): _description_. Defaults to None.
        num (int, optional): _description_. Defaults to 1.
        debug (bool, optional): _description_. Defaults to False.
        options (Text, optional): _description_. Defaults to None.
    """
    args_cmd_list = ['v6']
    dhcp_code_call_main(args_cmd_list)


if __name__ == '__main__':
    dhcp4_tool()
    # dhcp6_tool()
